# -*- coding: utf-8 -*-
"""
Block Korkine Zolotarev algorithm in Python.

..  moduleauthor:: Martin R.  Albrecht <martinralbrecht+fpylll@googlemail.com>

This module reimplements fplll's BKZ algorithm in Python.  It has feature parity with the C++
implementation in fplll's core.  Additionally, this implementation collects some additional
statistics.  Hence, it should provide a good basis for implementing variants of this algorithm.
"""
from __future__ import absolute_import

try:
    from time import process_time  # Python 3
except ImportError:
    from time import clock as process_time  # Python 2
from fpylll import IntegerMatrix, GSO, LLL
from fpylll import BKZ
from fpylll import Enumeration
from fpylll import EnumerationError
from fpylll.util import adjust_radius_to_gh_bound
from fpylll.tools.bkz_stats import dummy_tracer, normalize_tracer, Tracer


class BKZReduction(object):
    """
    An implementation of the BKZ algorithm in Python.

    This class has feature parity with the C++ implementation in fplll's core.  Additionally, this
    implementation collects some additional statistics.  Hence, it should provide a good basis for
    implementing variants of this algorithm.
    """

    def __init__(self, A):
        """Construct a new instance of the BKZ algorithm.

        :param A: an integer matrix, a GSO object or an LLL object

        """
        if isinstance(A, GSO.Mat):
            L = None
            M = A
            A = M.B
        elif isinstance(A, LLL.Reduction):
            L = A
            M = L.M
            A = M.B
        elif isinstance(A, IntegerMatrix):
            L = None
            M = None
            A = A
        else:
            raise TypeError("Matrix must be IntegerMatrix but got type '%s'" % type(A))

        if M is None and L is None:
            # run LLL first, but only if a matrix was passed
            LLL.reduction(A)

        self.A = A
        if M is None:
            self.M = GSO.Mat(A, flags=GSO.ROW_EXPO)
        else:
            self.M = M
        if L is None:
            self.lll_obj = LLL.Reduction(self.M, flags=LLL.DEFAULT)
        else:
            self.lll_obj = L

    def __call__(self, params, min_row=0, max_row=-1, tracer=False):
        """Run the BKZ algorithm with parameters `param`.

        :param params: BKZ parameters
        :param min_row: start processing in this row
        :param max_row: stop processing in this row (exclusive)
        :param tracer: see ``normalize_tracer`` for accepted values


        TESTS::

            >>> from fpylll import *
            >>> A = IntegerMatrix.random(60, "qary", k=30, q=127)
            >>> from fpylll.algorithms.bkz import BKZReduction
            >>> bkz = BKZReduction(A)
            >>> _ = bkz(BKZ.EasyParam(10), tracer=True); bkz.trace is None
            False
            >>> _ = bkz(BKZ.EasyParam(10), tracer=False); bkz.trace is None
            True

            >>> G = [4265, 7442,  5364,  5334,  6350, \
                     7442,  17726,  8388,  16002,  5461, \
                     5364,  8388,  6949,  5334,  9144, \
                     5334,  16002,  5334,  16129,  0, \
                     6350,  5461,  9144,  0,  16129]
            >>> G = IntegerMatrix.from_iterable(5, 5, G)
            >>> M = GSO.Mat(G, gram=True, update=True)
            >>> bkz = BKZReduction(M)
            >>> _ = bkz(BKZ.EasyParam(3))

        """

        tracer = normalize_tracer(tracer)

        try:
            label = params["name"]
        except KeyError:
            label = "bkz"

        if not isinstance(tracer, Tracer):
            tracer = tracer(
                self,
                root_label=label,
                verbosity=params.flags & BKZ.VERBOSE,
                start_clocks=True,
                max_depth=2,
            )

        if params.flags & BKZ.AUTO_ABORT:
            auto_abort = BKZ.AutoAbort(self.M, self.M.d)

        cputime_start = process_time()

        with tracer.context("lll"):
            self.lll_obj()

        i = 0
        while True:
            with tracer.context("tour", i, dump_gso=params.flags & BKZ.DUMP_GSO):
                clean = self.tour(params, min_row, max_row, tracer)
            i += 1
            if clean or params.block_size >= self.M.d:
                break
            if (params.flags & BKZ.AUTO_ABORT) and auto_abort.test_abort():
                break
            if (params.flags & BKZ.MAX_LOOPS) and i >= params.max_loops:
                break
            if (params.flags & BKZ.MAX_TIME) and process_time() - cputime_start >= params.max_time:
                break

        tracer.exit()
        try:
            self.trace = tracer.trace
        except AttributeError:
            self.trace = None
        return clean

    def tour(self, params, min_row=0, max_row=-1, tracer=dummy_tracer):
        """One BKZ loop over all indices.

        :param params: BKZ parameters
        :param min_row: start index ≥ 0
        :param max_row: last index ≤ n

        :returns: ``True`` if no change was made and ``False`` otherwise
        """
        if max_row == -1:
            max_row = self.M.d

        clean = True

        for kappa in range(min_row, max_row - 1):
            block_size = min(params.block_size, max_row - kappa)
            clean &= self.svp_reduction(kappa, block_size, params, tracer)

        self.lll_obj.size_reduction(max(0, max_row - 1), max_row, max(0, max_row - 2))
        return clean

    def svp_preprocessing(self, kappa, block_size, params, tracer):
        """Perform preprocessing for calling the SVP oracle

        :param kappa: current index
        :param params: BKZ parameters
        :param block_size: block size
        :param tracer: object for maintaining statistics

        :returns: ``True`` if no change was made and ``False`` otherwise

        .. note::

            ``block_size`` may be smaller than ``params.block_size`` for the last blocks.

        """
        clean = True

        lll_start = kappa if params.flags & BKZ.BOUNDED_LLL else 0
        with tracer.context("lll"):
            self.lll_obj(lll_start, lll_start, kappa + block_size)
            if self.lll_obj.nswaps > 0:
                clean = False

        return clean

    def svp_call(self, kappa, block_size, params, tracer=dummy_tracer):
        """Call SVP oracle

        :param kappa: current index
        :param params: BKZ parameters
        :param block_size: block size
        :param tracer: object for maintaining statistics

        :returns: Coordinates of SVP solution or ``None`` if none was found.

        ..  note::

            ``block_size`` may be smaller than ``params.block_size`` for the last blocks.
        """
        max_dist, expo = self.M.get_r_exp(kappa, kappa)
        delta_max_dist = self.lll_obj.delta * max_dist

        if params.flags & BKZ.GH_BND:
            root_det = self.M.get_root_det(kappa, kappa + block_size)
            max_dist, expo = adjust_radius_to_gh_bound(
                max_dist, expo, block_size, root_det, params.gh_factor
            )

        try:
            enum_obj = Enumeration(self.M)
            with tracer.context("enumeration", enum_obj=enum_obj, probability=1.0):
                max_dist, solution = enum_obj.enumerate(kappa, kappa + block_size, max_dist, expo)[
                    0
                ]

        except EnumerationError as msg:
            if params.flags & BKZ.GH_BND:
                return None
            else:
                raise EnumerationError(msg)

        if max_dist >= delta_max_dist * (1 << expo):
            return None
        else:
            return solution

    def svp_postprocessing(self, kappa, block_size, solution, tracer=dummy_tracer):
        """Insert SVP solution into basis. Note that this does not run LLL; instead,
           it resolves the linear dependencies internally.

        :param solution: coordinates of an SVP solution
        :param kappa: current index
        :param block_size: block size
        :param tracer: object for maintaining statistics

        :returns: ``True`` if no change was made and ``False`` otherwise

        ..  note :: postprocessing does not necessarily leave the GSO in a safe state.  You may
            need to call ``update_gso()`` afterwards.
        """
        if solution is None:
            return True

        # d = self.M.d
        # self.M.create_row()

        # with self.M.row_ops(d, d+1):
        #     for i in range(block_size):
        #         self.M.row_addmul(d, kappa + i, solution[i])

        # self.M.move_row(d, kappa)
        # with tracer.context("lll"):
        #     self.lll_obj(kappa, kappa, kappa + block_size + 1)
        # self.M.move_row(kappa + block_size, d)
        # self.M.remove_last_row()

        j_nz = None

        for i in range(block_size)[::-1]:
            if abs(solution[i]) == 1:
                j_nz = i
                break

        if len([x for x in solution if x]) == 1:
            self.M.move_row(kappa + j_nz, kappa)

        elif j_nz is not None:
            with self.M.row_ops(kappa + j_nz, kappa + j_nz + 1):
                for i in range(block_size):
                    if solution[i] and i != j_nz:
                        self.M.row_addmul(kappa + j_nz, kappa + i, solution[j_nz] * solution[i])

            self.M.move_row(kappa + j_nz, kappa)

        else:
            solution = list(solution)

            for i in range(block_size):
                if solution[i] < 0:
                    solution[i] = -solution[i]
                    self.M.negate_row(kappa + i)

            with self.M.row_ops(kappa, kappa + block_size):
                offset = 1
                while offset < block_size:
                    k = block_size - 1
                    while k - offset >= 0:
                        if solution[k] or solution[k - offset]:
                            if solution[k] < solution[k - offset]:
                                solution[k], solution[k - offset] = (
                                    solution[k - offset],
                                    solution[k],
                                )
                                self.M.swap_rows(kappa + k - offset, kappa + k)

                            while solution[k - offset]:
                                while solution[k - offset] <= solution[k]:
                                    solution[k] = solution[k] - solution[k - offset]
                                    self.M.row_addmul(kappa + k - offset, kappa + k, 1)

                                solution[k], solution[k - offset] = (
                                    solution[k - offset],
                                    solution[k],
                                )
                                self.M.swap_rows(kappa + k - offset, kappa + k)
                        k -= 2 * offset
                    offset *= 2

            self.M.move_row(kappa + block_size - 1, kappa)

        return False

    def svp_reduction(self, kappa, block_size, params, tracer=dummy_tracer):
        """Find shortest vector in projected lattice of dimension ``block_size`` and insert into
        current basis.

        :param kappa: current index
        :param params: BKZ parameters
        :param block_size: block size
        :param tracer: object for maintaining statistics

        :returns: ``True`` if no change was made and ``False`` otherwise
        """
        clean = True
        with tracer.context("preprocessing"):
            clean_pre = self.svp_preprocessing(kappa, block_size, params, tracer)
        clean &= clean_pre

        solution = self.svp_call(kappa, block_size, params, tracer)

        with tracer.context("postprocessing"):
            clean_post = self.svp_postprocessing(kappa, block_size, solution, tracer)
        clean &= clean_post

        self.lll_obj.size_reduction(0, kappa + 1)
        return clean



# -*- coding: utf-8 -*-
"""
Collecting traces from BKZ-like computations

.. moduleauthor:: Martin R. Albrecht <fplll-devel@googlegroups.com>

"""
from __future__ import print_function
from __future__ import absolute_import

import time

try:
    from time import process_time  # Python 3
except ImportError:
    from time import clock as process_time  # Python 2
import copy
from collections import OrderedDict
from math import log
from fpylll.tools.quality import basis_quality


def pretty_dict(d, keyword_width=None, round_bound=9999, suppress_length=128):
    """Return 'pretty' string representation of the dictionary ``d``.

    :param d: a dictionary
    :param keyword_width: width allocated for keywords
    :param round_bound: values beyond this bound are shown as `2^x`
    :param suppress_length: don't print arbitrary data with ``len(str(data)) > suppress_length``

    >>> from collections import OrderedDict
    >>> str(pretty_dict(OrderedDict([('d', 2), ('f', 0.1), ('large', 4097)])))
    '{"d":        2,  "f": 0.100000,  "large":     4097}'

    """
    s = []
    for k in d:

        v = d[k]

        if keyword_width:
            fmt = u'"%%%ds"' % keyword_width
            k = fmt % k
        else:
            k = '"%s"' % k

        if isinstance(v, int):
            if abs(v) > round_bound:
                s.append(u"%s: %8s" % (k, u"%s2^%.1f" % ("" if v > 0 else "-", log(abs(v), 2))))
            else:
                s.append(u"%s: %8d" % (k, v))
            continue
        elif not isinstance(v, float):
            try:
                v = float(v)
            except TypeError:
                if len(str(v)) <= suppress_length:
                    s.append(u"%s: %s" % (k, v))
                else:
                    s.append(u"%s: '...'" % (k,))
                continue

        if 0 <= v < 10.0:
            s.append(u"%s: %8.6f" % (k, v))
        elif -10 < v < 0:
            s.append(u"%s: %8.5f" % (k, v))
        elif abs(v) < round_bound:
            s.append(u"%s: %8.3f" % (k, v))
        else:
            s.append(u"%s: %8s" % (k, u"%s2^%.1f" % ("" if v > 0 else "-", log(abs(v), 2))))

    return u"{" + u",  ".join(s) + u"}"


class Accumulator(object):
    """
    An ``Accumulator`` collects observed facts about some random variable (e.g. running time).

    In particular,

        - minimum,

        - maximum,

        - mean and

        - variance

    are recorded::

        >>> v = Accumulator(1.0); v
        1.0
        >>> v += 2.0; v
        3.0
        >>> v = Accumulator(-5.4, repr="avg"); v
        -5.4
        >>> v += 0.2
        >>> v += 5.2; v
        0.0
        >>> v.min, v.max
        (-5.4, 5.2)

    """

    def __init__(self, value, repr="sum", count=True, bessel_correction=False):
        """
        Create a new instance.

        :param value: some initial value
        :param repr: how to represent the data: "min", "max", "avg", "sum" and "variance" are
            valid choices
        :param count: if ``True`` the provided value is considered as an observed datum, i.e. the
            counter is increased by one.
        :param bessel_correction: apply Bessel's correction to the variance.
        """

        self._min = value
        self._max = value
        self._sum = value
        self._sqr = value * value
        self._ctr = 1 if count else 0
        self._repr = repr
        self._bessel_correction = bessel_correction

    def add(self, value):
        """
        Add value to the accumulator.

        >>> v = Accumulator(10.0)
        >>> v.add(5.0)
        15.0

        :param value: some value
        :returns: itself

        """
        self._min = min(self._min, value)
        self._max = max(self._max, value)
        self._sum += value
        self._sqr += value * value
        self._ctr += 1
        return self

    @property
    def min(self):
        """
        >>> v = Accumulator(2.0)
        >>> v += 5.0
        >>> v.min
        2.0

        """
        return self._min

    @property
    def max(self):
        """
        >>> v = Accumulator(2.0)
        >>> v += 5.0
        >>> v.max
        5.0

        """
        return self._max

    @property
    def avg(self):
        """
        >>> v = Accumulator(2.0)
        >>> v += 5.0
        >>> v.avg
        3.5

        """
        return self._sum / self._ctr

    mean = avg

    @property
    def sum(self):
        """
        >>> v = Accumulator(2.0)
        >>> v += 5.0
        >>> v.sum
        7.0

        """
        return self._sum

    @property
    def variance(self):
        """
        >>> v = Accumulator(2.0)
        >>> v += 5.0
        >>> v.variance
        2.25

        """
        s = self._sqr / self._ctr - self.avg ** 2
        if self._bessel_correction:
            return self._ctr * (s / (self._ctr - 1))
        else:
            return s

    def __add__(self, other):
        """
        Addition semantics are:

        - ``stat + None`` returns ``stat``
        - ``stat + stat`` returns the sum of their underlying values
        - ``stat + value`` inserts ``value`` into ``stat``

        >>> v = Accumulator(2.0)
        >>> v + None
        2.0
        >>> v + v
        4.0
        >>> v + 3.0
        5.0

        """
        if other is None:
            return copy.copy(self)
        elif not isinstance(other, Accumulator):
            ret = copy.copy(self)
            return ret.add(other)
        else:
            if self._repr != other._repr:
                raise ValueError("%s != %s" % (self._repr, other._repr))
            ret = Accumulator(0)
            ret._min = min(self.min, other.min)
            ret._max = max(self.max, other.max)
            ret._sum = self._sum + other._sum
            ret._sqr = self._sqr + other._sqr
            ret._ctr = self._ctr + other._ctr
            ret._repr = self._repr
            return ret

    def __radd__(self, other):
        """
        Revert to normal addition.
        """
        return self + other

    def __sub__(self, other):
        """
        Return the difference of the two nodes reduced to floats.
        """
        return float(self) - float(other)

    def __float__(self):
        """
        Reduce this stats object down a float depending on strategy chosen in constructor.

        >>> v = Accumulator(2.0, "min"); v += 3.0; float(v)
        2.0
        >>> v = Accumulator(2.0, "max"); v += 3.0; float(v)
        3.0
        >>> v = Accumulator(2.0, "avg"); v += 3.0; float(v)
        2.5
        >>> v = Accumulator(2.0, "sum"); v += 3.0; float(v)
        5.0
        >>> v = Accumulator(2.0, "variance"); v += 3.0; float(v)
        0.25
        """
        return float(self.__getattribute__(self._repr))

    def __str__(self):
        """
        Reduce this stats object down a float depending on strategy chosen in constructor.

        >>> v = Accumulator(2.0, "min"); v += 3.0; str(v)
        '2.0'
        >>> v = Accumulator(2.0, "max"); v += 3.0; str(v)
        '3.0'
        >>> v = Accumulator(2.0, "avg"); v += 3.0; str(v)
        '2.5'
        >>> v = Accumulator(2.0, "sum"); v += 3.0; str(v)
        '5.0'
        >>> v = Accumulator(2.0, "variance"); v += 3.0; str(v)
        '0.25'
        """
        return str(self.__getattribute__(self._repr))

    __repr__ = __str__


class TraceContext(object):
    """
    A trace context collects data about an underlying process on entry/exit of particular parts of
    the code.
    """

    def __init__(self, tracer, *args, **kwds):
        """Create a new context for gathering statistics.

        :param tracer: a tracer object
        :param args: all args form a label for the trace context
        :param kwds: all kwds are considered auxiliary data

        """
        self.tracer = tracer
        self.what = args if len(args) > 1 else args[0]
        self.kwds = kwds

    def __enter__(self):
        """
        Call ``enter`` on trace object
        """
        self.tracer.enter(self.what, **self.kwds)

    def __exit__(self, exception_type, exception_value, exception_traceback):
        """
        Call ``exit`` on trace object
        """
        self.tracer.exit(**self.kwds)

        if exception_type is not None:
            return False


class Tracer(object):
    """
    A trace object is used to collect information about processes.

    This base class does nothing.
    """

    def __init__(self, instance, verbosity=False, max_depth=16):
        """
        Create a new tracer instance.

        :param instance: BKZ-like object instance
        :param verbosity: print information, integers ≥ 0 are also accepted
        :param max_depth: record up to this depth.

        """
        self.instance = instance
        self.verbosity = int(verbosity)
        self.max_depth = max_depth

    def context(self, *args, **kwds):
        """
        Return a new ``TraceCotext``.
        """
        return TraceContext(self, *args, **kwds)

    def enter(self, label, **kwds):
        """
        An implementation would implement this function which controls what happens when the context
        given by ``label`` is entered.
        """
        pass

    def exit(self, **kwds):
        """
        An implementation would implement this function which controls what happens when the context
        given by ``label`` is left.
        """
        pass

    def _pop(self):
        # NOTE: we handle ``max_depth`` by removing children when we exit.
        child = self.current
        self.current = self.current.parent
        if child.level > self.max_depth:
            self.current.del_child(child)


# use a dummy_trace whenever no tracing is required
dummy_tracer = Tracer(None)


class Node(object):
    """
    A simple tree implementation with labels and associated data.
    """

    def __init__(self, label, parent=None, data=None):
        """Create a new node.

        :param label: some label such as a string or a tuple
        :param parent: nodes know their parents
        :param data: nodes can have associated data which is a key-value store where typically the
            values are statistics
        """

        self.label = label
        if data is None:
            data = OrderedDict()
        self.data = OrderedDict(data)
        self.parent = parent
        self.children = []

    def add_child(self, child):
        """Add a child.

        :param child: a node
        :returns: the child

        """
        child.parent = self
        self.children.append(child)
        return child

    def del_child(self, child):
        """

        >>> root = Node("root")
        >>> c1 = root.child("child1")
        >>> c2 = root.child("child2")
        >>> root.children
        [{"child1": {}}, {"child2": {}}]
        >>> root.del_child(c1)
        >>> root.children
        [{"child2": {}}]

        """
        self.children.remove(child)

    def child(self, label):
        """
        If node has a child labelled ``label`` return it, otherwise add a new child.

        :param label: a label
        :returns: a node

        >>> root = Node("root")
        >>> c1 = root.child("child"); c1
        {"child": {}}
        >>> c2 = root.child("child"); c2
        {"child": {}}
        >>> c1 is c2
        True

        """
        for child in self.children:
            if child.label == label:
                return child

        return self.add_child(Node(label))

    def __str__(self):
        """
        >>> from collections import OrderedDict
        >>> str(Node("root", data=OrderedDict([('a',1), ('b', 2)])))
        '{"root": {"a":        1,  "b":        2}}'
        """
        return u'{"%s": %s}' % (self.label, pretty_dict(self.data))

    __repr__ = __str__

    def report(self, indentation=0, depth=None):
        """
        Return detailed string representation of this tree.

        :param indentation: add spaces to the left of the string representation
        :param depth: stop at this depth

        >>> root = Node("root")
        >>> c1 = root.child(("child",1))
        >>> c2 = root.child(("child",2))
        >>> c3 = c1.child(("child", 3))
        >>> c1.data["a"] = 100.0
        >>> c3.data["a"] = 4097

        >>> print(root.report())
        {"root": {}}
          {"('child', 1)": {"a":  100.000}}
            {"('child', 3)": {"a":     4097}}
          {"('child', 2)": {}}

        >>> print(root.report(indentation=2, depth=1))
          {"root": {}}
            {"('child', 1)": {"a":  100.000}}
            {"('child', 2)": {}}
        """
        s = [" " * indentation + str(self)]
        if depth is None or depth > 0:
            for child in self.children:
                depth = None if depth is None else depth - 1
                s.append(child.report(indentation + 2, depth=depth))
        return "\n".join(s)

    def sum(self, tag, include_self=True, raise_keyerror=False, label=None):
        """
        Return sum over all items tagged ``tag`` in associated data within this tree.

        :param tag: a string
        :param include_self: include data in this node
        :param raise_keyerror: if a node does not have an item tagged with ``tag`` raise a
            ``KeyError``
        :param label: filter by ``label``


        >>> root = Node("root")
        >>> c1 = root.child(("child",1))
        >>> c2 = root.child(("child",2))
        >>> c3 = c1.child(("child", 3))
        >>> c1.data["a"] = 100.0
        >>> c3.data["a"] = 4097

        >>> root.sum("a")
        4197.0

        >>> root.sum("a", label=("child",3))
        4097

        >>> root.sum("a", label=("child",2))
        0

        >>> root.sum("a", label=("child",2), raise_keyerror=True)
        Traceback (most recent call last):
        ...
        KeyError: 'a'

        """
        if include_self and (label is None or self.label == label):
            if raise_keyerror:
                r = self.data[tag]
            else:
                r = self.data.get(tag, 0)
        else:
            r = 0
        for child in self.children:
            r = r + child.sum(tag, include_self=True, raise_keyerror=raise_keyerror, label=label)
        return r

    def find(self, label, raise_keyerror=False):
        """
        Find the first child node matching label in a breadth-first search.

        :param label: a label
        :param raise_keyerror: raise a ``KeyError`` if ``label`` was not found.
        """
        for child in self.children:
            if child.label == label:
                return child
        for child in self.children:
            try:
                return child.find(label, raise_keyerror)
            except KeyError:
                pass

        if raise_keyerror:
            raise KeyError("Label '%s' not present in '%s" % (label, self))
        else:
            return None

    def find_all(self, label):
        """
        Find all nodes labelled ``label``

        :param label: a label

        """
        r = []
        if self.label == label:
            r.append(self)
        if isinstance(self.label, tuple) and self.label[0] == label:
            r.append(self)
        for child in self.children:
            r.extend(child.find_all(label))
        return tuple(r)

    def __iter__(self):
        """
        Depth-first iterate over all subnodes (including self)

        ::

            >>> root = Node("root")
            >>> c1 = root.child(("child",1))
            >>> c2 = root.child(("child",2))
            >>> c3 = c1.child(("child", 3))
            >>> c1.data["a"] = 100.0
            >>> c3.data["a"] = 4097
            >>> list(root)
            [{"root": {}}, {"('child', 1)": {"a":  100.000}}, {"('child', 3)": {"a":     4097}}, {"('child', 2)": {}}]

        """

        yield self
        for child in self.children:
            for c in iter(child):
                yield c

    def merge(self, node):
        """
        Merge tree ``node`` into self.

        .. note :: The label of ``node`` is ignored.
        """

        for k, v in node.data.iteritems():
            if k in self.data:
                self.data[k] += v
            else:
                self.data[k] = v

        for child in node.children:
            self.child(child.label).merge(child)

    def get(self, label):
        """Return first child node with label ``label``

        :param label: label

        >>> root = Node("root")
        >>> _ = root.child("bar")
        >>> c1 = root.child(("foo",0))
        >>> c2 = root.child(("foo",3))
        >>> c3 = c1.child(("foo", 3))
        >>> c1.data["a"] = 100.0
        >>> c3.data["a"] = 4097
        >>> root.get("bar")
        {"bar": {}}

        >>> root.get("foo")
        ({"('foo', 0)": {"a":  100.000}}, {"('foo', 3)": {}})

        >>> root.get("foo")[0]
        {"('foo', 0)": {"a":  100.000}}
        >>> root.get("foo")[1]
        {"('foo', 3)": {}}

        """
        r = []
        for child in self.children:
            if child.label == label:
                return child
            if isinstance(child.label, tuple) and child.label[0] == label:
                r.append(child)
        if r:
            return tuple(r)
        else:
            raise AttributeError("'Node' object has no attribute '%s'" % (label))

    def __getitem__(self, tag):
        """Return associated data tagged ``tag```

        :param tag: Some tag

        >>> root = Node("root", data={"foo": 1})
        >>> c1 = root.child("foo")
        >>> root["foo"]
        1

        """
        return self.data[tag]

    @property
    def level(self):
        """
        Return level of this node, i.e. how many steps it takes to reach a node with parent
        ``None``.

        >>> root = Node("root")
        >>> _ = root.child("bar")
        >>> c1 = root.child(("foo",0))
        >>> c2 = root.child(("foo",3))
        >>> c3 = c1.child(("foo", 3))
        >>> root.level
        0
        >>> c1.level
        1
        >>> c3.level
        2

        """
        node, level = self, 0
        while node.parent is not None:
            level += 1
            node = node.parent
        return level

    def __sub__(self, rhs):
        """
        Return tree that contains the difference of this node and the other.

        The semantics are as follows:

        - For all data in this node the matching data item in ``rhs`` is subtracted.
        - If the data is missing in ``rhs`` it is assumed to be zero.
        - For all children of this node this function is called recursively.
        - If ``rhs`` does not have an immediate child node with a matching label, those children are skipped.
        """

        if not isinstance(rhs, Node):
            raise ValueError("Expected node but got '%s'" % type(rhs))
        diff = Node(self.label)
        for k in self.data:
            diff.data[k] = self.data[k] - rhs.data.get(k, 0)

        for lchild in self.children:
            for rchild in rhs.children:
                if lchild.label == rchild.label:
                    diff.children.append(lchild - rchild)
                    break
            else:
                print("Skipping missing node '%s'" % lchild.label)
        return diff

    def copy(self, deepcopy=True):
        """
        Return a (deep)copy of this node.

        :param deepcopy: If ``False`` child nodes and data dictionaries are not copied, this is
            usually not what the user wants.
        """
        if deepcopy:
            return copy.deepcopy(self)
        else:
            return copy.copy(self)

    def accumulate(self, key, filter=lambda node: True, repr="avg"):
        """
        Return accumulator value for all occurrences of ``key``::

            >>> root = Node("root")
            >>> c1 = root.child(("child",1))
            >>> c2 = root.child(("child",2))
            >>> c3 = c1.child(("child", 3))
            >>> c1.data["a"] = 100.0
            >>> c3.data["a"] = 4097
            >>> root.accumulate("a").sum
            4197.0
            >>> root.accumulate("a", filter=lambda node: node.label == ("child", 3)).sum
            4097


        :param key: Dictionary key to some ``data`` attribute
        :param filter: Callable that should return ``True`` for nodes that ought to be considered.
        :param repr: Representation of accumulator

        """
        acc = Accumulator(0, repr=repr, count=False)
        for node in iter(self):
            if filter(node) and key in node.data:
                acc += node.data[key]
        return acc


class TimeTreeTracer(Tracer):
    """
    Collect CPU and wall time for every context visited, creating a tree structure along the way.
    """

    def __init__(
        self,
        instance,
        verbosity=False,
        root_label="root",
        start_clocks=False,
        max_depth=1024,
    ):
        """
        Create a new tracer instance.

        :param instance: BKZ-like object instance
        :param verbosity: print information, integers ≥ 0 are also accepted
        :param root_label: label to give to root node
        :param start_clocks: start tracking time for the root node immediately
        :param max_depth: record up to this depth.

        """

        Tracer.__init__(self, instance, verbosity, max_depth)
        self.trace = Node(root_label)
        self.current = self.trace
        if start_clocks:
            self.reenter()

    def enter(self, label, **kwds):
        """Enter new context with label

        :param label: label

        """
        self.current = self.current.child(label)
        self.reenter()

    def reenter(self, **kwds):
        """Reenter current context, i.e. restart clocks"""

        if self.current is None:
            # we exited the root node
            self.current = self.trace
        node = self.current
        node.data["cputime"] = node.data.get("cputime", 0) + Accumulator(
            -process_time(), repr="sum", count=False
        )
        node.data["walltime"] = node.data.get("walltime", 0) + Accumulator(
            -time.time(), repr="sum", count=False
        )

    def exit(self, **kwds):
        """
        Leave context, record time spent.

        :param label: ignored

        .. note :: If verbosity ≥ to the current level, also print the current node.

        """
        node = self.current

        node.data["cputime"] += process_time()
        node.data["walltime"] += time.time()

        if self.verbosity and self.verbosity >= self.current.level:
            print(self.current)

        self._pop()


class BKZTreeTracer(Tracer):
    """
    Default tracer for BKZ-like algorithms.
    """

    def __init__(
        self, instance, verbosity=False, root_label="bkz", start_clocks=False, max_depth=16
    ):
        """
        Create a new tracer instance.

        :param instance: BKZ-like object instance
        :param verbosity: print information, integers ≥ 0 are also accepted
        :param root_label: label to give to root node
        :param start_clocks: start tracking time for the root node immediately
        :param max_depth: record up to this depth.

        TESTS::

            >>> from fpylll.tools.bkz_stats import BKZTreeTracer
            >>> tracer = BKZTreeTracer(None)
            >>> for i in range(3): tracer.enter("level-%d"%i)
            >>> for i in range(3): tracer.exit()
            >>> "level-2" in tracer.trace.report()
            True
            >>> tracer = BKZTreeTracer(None, max_depth=2)
            >>> for i in range(3): tracer.enter("level-%d"%i)
            >>> for i in range(3): tracer.exit()
            >>> "level-2" in tracer.trace.report()
            False

        """

        Tracer.__init__(self, instance, verbosity, max_depth)
        self.trace = Node(root_label)
        self.current = self.trace
        if start_clocks:
            self.reenter()

    def enter(self, label, **kwds):
        """Enter new context with label

        :param label: label

        """
        self.current = self.current.child(label)
        self.reenter()

    def reenter(self, **kwds):
        """Reenter current context, i.e. restart clocks"""

        node = self.current
        node.data["cputime"] = node.data.get("cputime", 0) + Accumulator(
            -process_time(), repr="sum", count=False
        )
        node.data["walltime"] = node.data.get("walltime", 0) + Accumulator(
            -time.time(), repr="sum", count=False
        )

    def exit(self, **kwds):  # noqa, shut up linter about this function being too complex
        """
        By default CPU and wall time are recorded.  More information is recorded for "enumeration"
        and "tour" labels.  When the label is a tour then the status is printed if verbosity > 0.
        """
        node = self.current
        label = node.label

        node.data["cputime"] += process_time()
        node.data["walltime"] += time.time()

        if kwds.get("dump_gso", False):
            node.data["r"] = node.data.get("r", []) + [self.instance.M.r()]

        if label == "enumeration":
            full = kwds.get("full", True)
            if full:
                try:
                    node.data["#enum"] = Accumulator(
                        kwds["enum_obj"].get_nodes(), repr="sum"
                    ) + node.data.get(
                        "#enum", None
                    )  # noqa
                except KeyError:
                    pass
                try:
                    node.data["%"] = Accumulator(kwds["probability"], repr="avg") + node.data.get(
                        "%", None
                    )
                except KeyError:
                    pass

        if label[0] == "tour":
            data = basis_quality(self.instance.M)
            for k, v in data.items():
                if k == "/":
                    node.data[k] = Accumulator(v, repr="max")
                else:
                    node.data[k] = Accumulator(v, repr="min")

        if self.verbosity and label[0] == "tour":
            report = OrderedDict()
            report["i"] = label[1]
            report["cputime"] = node["cputime"]
            report["walltime"] = node["walltime"]
            try:
                report["preproc"] = node.find("preprocessing", True)["cputime"]
            except KeyError:
                pass
            try:
                report["svp"] = node.find("enumeration", True)["cputime"]
            except KeyError:
                pass
            report["#enum"] = node.sum("#enum")
            report["lll"] = node.sum("cputime", label="lll")
            try:
                report["pruner"] = node.find("pruner", True)["cputime"]
            except KeyError:
                pass
            report["r_0"] = node["r_0"]
            report["/"] = node["/"]

            print(pretty_dict(report))

        self._pop()


def normalize_tracer(tracer):
    """
    Normalize tracer inputs for convenience.

    :param tracer:  ``True`` for ``BKZTreeTracer``, ``False`` for ``dummy_tracer``
                     or any other value for custom tracer.

    EXAMPLE::

        >>> from fpylll.tools.bkz_stats import normalize_tracer, BKZTreeTracer, dummy_tracer
        >>> normalize_tracer(True) == BKZTreeTracer
        True
        >>> normalize_tracer(False) == dummy_tracer
        True
        >>> normalize_tracer(BKZTreeTracer) == BKZTreeTracer
        True

    """
    if tracer is True:
        return BKZTreeTracer
    elif tracer is False:
        return dummy_tracer
    else:
        return tracer



# -*- coding: utf-8 -*-
"""
This file implements two BKZ simulation algorithms as proposed in

- Chen, Y., & Nguyen, P. Q. (2011). BKZ 2.0: better lattice security estimates.
  In D. H. Lee, & X.  Wang, ASIACRYPT~2011 (pp. 1–20). : Springer, Heidelberg.

- Bai, S., & Stehlé, D. & Wen, W. (2018). Measuring, simulating and exploiting
  the head concavity phenomenon in BKZ. In T. Peyrin, & S. Galbraith,
  ASIACRYPT~2018 : Springer, Heidelberg.

.. moduleauthor:: Michael Walter <fplll-devel@googlegroups.com> (2014)
.. moduleauthor:: Martin R. Albrecht <fplll-devel@googlegroups.com> (2018)
.. moduleauthor:: Shi Bai <fplll-devel@googlegroups.com> (2020)
.. moduleauthor:: Fernando Virdia <fernando.virdia.2016@rhul.ac.uk> (2020)
.. moduleauthor:: Ludo Pulles <lnp@cwi.nl> (2024)

"""
from copy import copy
from math import log, sqrt, lgamma, pi
from collections import OrderedDict
import random

from fpylll.tools.quality import basis_quality
from fpylll.tools.bkz_stats import pretty_dict
from fpylll.fplll.bkz import BKZ
from fpylll.fplll.integer_matrix import IntegerMatrix
from fpylll.fplll.gso import MatGSO, GSO
from fpylll import FPLLL

# Line 2, Algorithm 2 of [CN11]:
# Average log2(||b_k*||) of an HKZ-reduced random unit-volume 45-dim lattice.
rk = (
    0.789527997160000,
    0.780003183804613,
    0.750872218594458,
    0.706520454592593,
    0.696345241018901,
    0.660533841808400,
    0.626274718790505,
    0.581480717333169,
    0.553171463433503,
    0.520811087419712,
    0.487994338534253,
    0.459541470573431,
    0.414638319529319,
    0.392811729940846,
    0.339090376264829,
    0.306561491936042,
    0.276041187709516,
    0.236698863270441,
    0.196186341673080,
    0.161214212092249,
    0.110895134828114,
    0.0678261623920553,
    0.0272807162335610,
    -0.0234609979600137,
    -0.0320527224746912,
    -0.0940331032784437,
    -0.129109087817554,
    -0.176965384290173,
    -0.209405754915959,
    -0.265867993276493,
    -0.299031324494802,
    -0.349338597048432,
    -0.380428160303508,
    -0.427399405474537,
    -0.474944677694975,
    -0.530140672818150,
    -0.561625221138784,
    -0.612008793872032,
    -0.669011014635905,
    -0.713766731570930,
    -0.754041787011810,
    -0.808609696192079,
    -0.859933249032210,
    -0.884479963601658,
    -0.886666930030433,
)


def _lg_gh(d):
    """
    Return the logarithm (base 2) of the Gaussian Heuristic, GH(d).
    Here GH(d) is the expected length of the shortest nonzero vector in a
    random lattice of dimension d.
    :param d: dimension
    """
    if d <= 45:
        return rk[-d] - sum(rk[-d:]) / d
    log_gh = lgamma(d / 2.0 + 1) * (1.0 / d) - log(sqrt(pi))
    return log_gh / log(2.0)


def log_simulate(basis_profile, beta, max_loops=False, verbose=False):
    """
    Simulate the evolution of the basis profile when running multiple tours of
    BKZ-beta, using the simulator Algorithm 1 in [CN11].

    :param basis_profile: log2(norm) of the GSO vectors of the basis.
    :param max_loops: maximum number of full tours in BKZ to perform.
                      BKZ terminates earlier if no progress is made in a tour.
    :param beta: block size for BKZ.
    :returns: tuple containing:
              1. the basis profile after BKZ finished,
              2. number of tours performed.
    """
    profile, n = copy(basis_profile), len(basis_profile)
    assert 2 <= beta <= n

    lg_ghs = [0] + [_lg_gh(i) for i in range(1, beta + 1)]
    if not max_loops:
        max_loops = n

    for j in range(max_loops):
        lg_volume = sum(profile[:beta])  # lg_volume = sum(profile[k:f])
        updated = False
        for k in range(0, n - 1):
            f = min(k + beta, n)  # end index (excl.) of local block [k, f)
            d = f - k  # dimension of local block
            svp_sol = lg_volume/d + lg_ghs[d]
            if updated or svp_sol < profile[k]:
                profile[k] = svp_sol
                updated = True

            lg_volume -= profile[k]  # Remove ||b_k*|| from the sliding window
            if f < n:
                lg_volume += profile[f]  # Add ||b_f*|| to the sliding window
        profile[-1] = lg_volume  # sum(profile) = lg(det L)
        if not updated:
            return profile, j + 1  # Early termination (unlikely)
        if verbose:
            stats = {'i': j} | basis_quality([2.0**(2 * x) for x in profile])
            print(pretty_dict(OrderedDict(stats)))

    return profile, max_loops


def log_simulate_prob(basis_profile, block_size, prng_seed=0xdeadbeef,
                      max_loops=False, verbose=False):
    """
    Simulate the evolution of the basis profile when running multiple tours of
    BKZ-beta, using the simulator by Bai, Stehlé and Wen [BSW18].

    :param basis_profile: log2(norm) of the GSO vectors of the basis.
    :param max_loops: maximum number of full tours in BKZ to perform.
                      BKZ terminates earlier if no progress is made in a tour.
    :param beta: block size for BKZ.
    :returns: tuple containing:
              1. the basis profile after BKZ finished,
              2. number of tours performed.
    """
    if block_size <= 2:
        raise ValueError("The BSW18 simulator requires block size >= 3.")

    # fix PRNG seed
    random.seed(prng_seed if prng_seed else FPLLL.randint(0, 2**32 - 1))

    d = len(basis_profile)

    r1 = copy(basis_profile)
    r2 = copy(basis_profile)
    lg_ghs = [0] + [_lg_gh(i) for i in range(1, block_size + 1)]

    if not max_loops:
        max_loops = d

    t0 = [True for _ in range(d)]
    for i in range(max_loops):
        t1 = [False for _ in range(d)]
        for k in range(d - min(45, block_size)):
            beta = min(block_size, d - k)
            f = k + beta
            phi = False
            for kp in range(k, f):
                phi |= t0[kp]
            lg_volume = sum(r1[:f]) - sum(r2[:k])
            if phi:
                x = random.expovariate(.5)
                lma = (log(x, 2) + lg_volume) / beta + lg_ghs[beta]
                if lma < r1[k]:
                    r2[k] = lma
                    r2[k+1] = r1[k] + log(sqrt(1 - 1./beta), 2)
                    dec = (r1[k] - lma) + (r1[k+1] - r2[k+1])
                    for j in range(k+2, f):
                        r2[j] = r1[j] + dec/(beta - 2.)
                        t1[j] = True
                    phi = False

            for j in range(k, f):
                r1[j] = r2[j]

        if not any(t1):
            return r1, i + 1  # early termination

        # The last block looks HKZ-reduced, similar to [CN11].
        beta = min(45, block_size)
        lg_volume = sum(r1) - sum(r2[:-beta]) - sum(rk[-beta:])
        for k in range(d - beta, d):
            r2[k] = lg_volume / beta + rk[-(d - k)]
            t1[k] = True

        if r1 == r2:
            return r1, i + 1  # early termination
        r1 = copy(r2)
        t0 = copy(t1)

        if verbose:
            stats = {'i': i} | basis_quality([2.0**(2 * x) for x in r1])
            print(pretty_dict(OrderedDict(stats)))

    return r1, max_loops


def _extract_log_norms(r):
    if isinstance(r, IntegerMatrix):
        r = GSO.Mat(r)
    elif isinstance(r, MatGSO):
        r.update_gso()
        r = r.r()
    else:
        for ri in r:
            if ri <= 0:
                raise ValueError("squared norms in r should be positive")

    # code uses log2 of norms, FPLLL uses squared norms
    return list(map(lambda x: log(x, 2) / 2.0, r))


def simulate(r, param):
    """
    BKZ simulation algorithm as proposed by Chen and Nguyen [CN11].  This
    version terminates when no substantial progress is made anymore or when
    ``max_loops`` tours were simulated. If no ``max_loops`` is given, at most
    ``d`` tours are performed, where ``d`` is the dimension of the lattice.
    :param r: squared norms of the GSO vectors of the basis.
    :param param: BKZ parameters (block_size, max_loops, verbose)
    :returns: tuple with:
              1. the reduced squared norms of the GSO vectors of the basis,
              2. and the number of BKZ tours simulated.
    EXAMPLE::
        >>> from fpylll import IntegerMatrix, GSO, LLL, FPLLL, BKZ
        >>> FPLLL.set_random_seed(1337)
        >>> A = LLL.reduction(IntegerMatrix.random(100, "qary", bits=30, k=50))
        >>> M = GSO.Mat(A)
        >>> from fpylll.tools.bkz_simulator import simulate
        >>> _ = simulate(M, BKZ.Param(block_size=2, max_loops=4, flags=BKZ.VERBOSE))
        {"i":        0,  "r_0":   2^35.2,  "r_0/gh": 7.720730,  "rhf": 1.019532,  "/": -0.08309,  "hv/hv": 2.831284}
        {"i":        1,  "r_0":   2^35.2,  "r_0/gh": 7.513168,  "rhf": 1.019393,  "/": -0.08300,  "hv/hv": 2.828248}
        {"i":        2,  "r_0":   2^35.2,  "r_0/gh": 7.492349,  "rhf": 1.019379,  "/": -0.08291,  "hv/hv": 2.825635}
        {"i":        3,  "r_0":   2^35.1,  "r_0/gh": 7.427548,  "rhf": 1.019335,  "/": -0.08282,  "hv/hv": 2.822987}
        >>> _ = simulate(M, BKZ.Param(block_size=40, max_loops=4, flags=BKZ.VERBOSE))
        {"i":        0,  "r_0":   2^34.7,  "r_0/gh": 5.547855,  "rhf": 1.017849,  "/": -0.06928,  "hv/hv": 2.406481}
        {"i":        1,  "r_0":   2^34.2,  "r_0/gh": 3.894188,  "rhf": 1.016049,  "/": -0.06136,  "hv/hv": 2.150078}
        {"i":        2,  "r_0":   2^33.8,  "r_0/gh": 2.949459,  "rhf": 1.014638,  "/": -0.05735,  "hv/hv": 2.044402}
        {"i":        3,  "r_0":   2^33.6,  "r_0/gh": 2.574565,  "rhf": 1.013949,  "/": -0.05556,  "hv/hv": 1.999163}
    """
    profile, num_loops = log_simulate(
        _extract_log_norms(r), param.block_size, param.max_loops,
        param.flags & BKZ.VERBOSE,
    )
    return [2.0**(2 * x) for x in profile], num_loops


def simulate_prob(r, param, prng_seed=0xdeadbeef):
    """
    BKZ simulation algorithm as proposed by Bai, Stehlé and Wen [BSW18].  This
    version terminates when no substantial progress is made anymore or when
    ``max_loops`` tours were simulated. If no ``max_loops`` is given, at most
    ``d`` tours are performed, where ``d`` is the dimension of the lattice.
    :param r: squared norms of the GSO vectors of the basis.
    :param param: BKZ parameters (block_size, max_loops, verbose)
    :returns: tuple with:
              1. the reduced squared norms of the GSO vectors of the basis,
              2. and the number of BKZ tours simulated.
    EXAMPLE::
        >>> from fpylll import IntegerMatrix, GSO, LLL, FPLLL, BKZ
        >>> FPLLL.set_random_seed(1337)
        >>> A = LLL.reduction(IntegerMatrix.random(100, "qary", bits=30, k=50))
        >>> M = GSO.Mat(A)
        >>> from fpylll.tools.bkz_simulator import simulate_prob
        >>> _ = simulate_prob(M, BKZ.Param(block_size=40, max_loops=4, flags=BKZ.VERBOSE))
        {"i":        0,  "r_0":   2^34.5,  "r_0/gh": 4.714937,  "rhf": 1.017021,  "/": -0.06936,  "hv/hv": 2.410445}
        {"i":        1,  "r_0":   2^34.2,  "r_0/gh": 3.874259,  "rhf": 1.016023,  "/": -0.06189,  "hv/hv": 2.162205}
        {"i":        2,  "r_0":   2^33.8,  "r_0/gh": 2.996068,  "rhf": 1.014718,  "/": -0.05798,  "hv/hv": 2.056934}
        {"i":        3,  "r_0":   2^33.7,  "r_0/gh": 2.773499,  "rhf": 1.014326,  "/": -0.05598,  "hv/hv": 2.012050}
    """
    profile, num_loops = log_simulate_prob(
        _extract_log_norms(r), param.block_size, prng_seed, param.max_loops,
        param.flags & BKZ.VERBOSE,
    )
    return [2.0**(2 * x) for x in profile], num_loops


def averaged_simulate_prob(r, param, tries=10):
    """
    This wrapper calls the [BSW18] probabilistic BKZ simulator with different
    PRNG seeds, and returns the average output.
    Note: exp(E[ log(X_i) ]) is reported where X_i is the squared norm of the
    i-th GSO vector.
    :param r: squared norms of the GSO vectors of the basis.
    :param param: BKZ parameters (block_size, max_loops, verbose)
    :param tries: number of iterations to average over. Default: 10
    :returns: tuple with:
              1. averaged reduced squared norms of GSO vectors of the basis,
              2. the averaged number of BKZ tours simulated.
    EXAMPLE::
        >>> from fpylll import IntegerMatrix, GSO, LLL, FPLLL, BKZ
        >>> from fpylll.tools.bkz_simulator import averaged_simulate_prob
        >>> FPLLL.set_random_seed(1337)
        >>> A = LLL.reduction(IntegerMatrix.random(100, "qary", bits=30, k=50))
        >>> M = GSO.Mat(A)
        >>> _ = averaged_simulate_prob(M, BKZ.Param(block_size=40, max_loops=4))
        >>> print(_[0][:3])
        [13371442256.252..., 12239031147.433..., 12256303707.863...]
    """
    if tries < 1:
        raise ValueError("Need to average over positive number of tries.")
    profile = _extract_log_norms(r)

    profiles, loops = zip(*[log_simulate_prob(
        profile, param.block_size, i + 1, param.max_loops,
        param.flags & BKZ.VERBOSE
    ) for i in range(tries)])

    avg_profile = [sum(x) / len(x) for x in zip(*profiles)]
    avg_loops = sum(loops) / tries
    return [2.0**(2 * x) for x in avg_profile], avg_loops



# -*- coding: utf-8 -*-
"""
Babai's Nearest Plane algorithm


..  moduleauthor:: Martin R.  Albrecht <martinralbrecht+fpylll@googlemail.com>

"""

from fpylll import IntegerMatrix, LLL
from math import isqrt


def babai(B, t, *args, **kwargs):
    """
    Run Babai's Nearest Plane algorithm by running LLL.

    :param B: Input lattice basis.
    :param target: Target point (∈ ZZ^n)
    :param args: Passed onto LLL.reduction
    :param kwargs: Passed onto LLL.reduction
    :returns coordinates of the solution vector:

    This implementation is more numerically stable compared to the one offered by `MatGSO.babai()`.
    On the other hand, this implementation will only accept tatgets with Integer coefficients.

    EXAMPLE::

       >>> from fpylll import *
       >>> n = 10
       >>> B = IntegerMatrix(n, n + 1)
       >>> B.randomize("intrel", bits=100)
       >>> v_opt = B.multiply_left([1,0,1,0,1,1,0,0,1,1])
       >>> s = v_opt[0] # s = <a, x>, where a is vector of knapsack values.
       >>> t = [s] + (n * [0])
       >>> v = CVP.babai(B, t)
       >>> v[0] == t[0]
       True
       >>> v[1:]
       (1, 0, 1, 0, 1, 1, 0, 0, 1, 1)
       >>> _ = LLL.reduction(B)
       >>> v == CVP.closest_vector(B, t)
       True

       >>> from fpylll import *
       >>> A = [428812188057600, 1, 0, 409148314550272, 0, 1]
       >>> A = IntegerMatrix.from_iterable(2, 3, A)
       >>> CVP.babai(A, (292633475057909760, 256, 256))
       (292633475057909760, 296, 405)

    """
    A = IntegerMatrix(B.nrows + 1, B.ncols + 1)
    for i in range(B.nrows):
        for j in range(B.ncols):
            A[i, j] = B[i, j]

    # make sure the input is LLL reduced before reading the norm of the last vector
    LLL.reduction(A, *args, **kwargs)
    # zero vector at the end
    A.swap_rows(0, B.nrows)

    for j in range(B.ncols):
        A[-1, j] = t[j]
    # precise norm, +1 to make sure it's not too small, too big doesn't matter
    A[-1, -1] = isqrt(sum(x**2 for x in A[-2])) + 1

    LLL.reduction(A, *args, **kwargs)  # now call LLL to run Babai

    # HACK: LLL might have done some swaps, but it shouldn't have!
    for row in range(A.nrows):
        if A[row, -1] != 0:
            break

    v = [0] * len(t)
    if A[row, -1] > 0:
        for i in range(len(t)):
            v[i] = t[i] - A[row][i]
    else:
        for i in range(len(t)):
            v[i] = t[i] + A[row][i]

    return tuple(v)



# -*- coding: utf-8 -*-
"""

"""

from math import log, exp
from collections import OrderedDict
from fpylll.util import gaussian_heuristic


def get_current_slope(r, start_row=0, stop_row=-1):
    """
    A Python re-implementation of ``MatGSO.get_current_slope``.

        >>> from fpylll import IntegerMatrix, GSO, LLL, FPLLL
        >>> FPLLL.set_random_seed(1337)
        >>> A = IntegerMatrix.random(100, "qary", bits=30, k=50)
        >>> _ = LLL.reduction(A)
        >>> M = GSO.Mat(A); _ = M.update_gso()
        >>> from fpylll.tools.quality import get_current_slope
        >>> M.get_current_slope(0, 100)  # doctest: +ELLIPSIS
        -0.083173398...
        >>> get_current_slope(M.r(), 0, 100) # doctest: +ELLIPSIS
        -0.083173398...

    """
    x = [log(r[i]) for i in range(start_row, stop_row)]
    n = stop_row - start_row
    i_mean = (n - 1) * 0.5 + start_row
    x_mean = sum(x)/n
    v1, v2 = 0.0, 0.0

    for i in range(start_row, stop_row):
        v1 += (i - i_mean) * (x[i] - x_mean)
        v2 += (i - i_mean) * (i - i_mean)
    return v1 / v2


def basis_quality(M):
    r"""
    Return a dictionary with various expressions of quality of the basis corresponding to ``M``.

    Let `|b_i^*|` be the norm of the `i`-th Gram-Schmidt vector.  Let `Λ` be the lattice spanned by
    the basis of dimension `d`.

        - ``r_0`` - `|b_0|^2`

        - ``\`` - the slope of `\log(|b_i^*|)`

        - ``rhf`` - the root-Hermite factor `|b_0|/\Vol(Λ)^{1/d}` also written as `\delta_0`

        - ``hv/hv`` - the dth-root of the fraction of the first and second half-volumes, i.e. the
          dth-root of `∏_{i=0}^{d/2-1} |b_i|/∏_{i=d/2}^{d-1} |b_i|`.  If `d` is odd, the length
          `|b_{d//2}|` is ignored.

        - ``r_0/gh`` - `|b_0|/GH` where `GH = Γ(d/2+1)^{1/d}/π^{1/2} ⋅ \Vol(Λ)^{1/d}` is the Gaussian
          Heuristic for the shortest vector.

    :param M: A MatGSO object or a vector of squared Gram-Schmidt norms.

    Example:

        >>> from fpylll import IntegerMatrix, GSO, LLL, FPLLL
        >>> FPLLL.set_random_seed(1337)
        >>> A = IntegerMatrix.random(100, "qary", bits=30, k=50)
        >>> _ = LLL.reduction(A)
        >>> M = GSO.Mat(A); _ = M.update_gso()

        >>> from fpylll.tools.quality import basis_quality
        >>> from fpylll.tools.bkz_stats import pretty_dict
        >>> str(pretty_dict(basis_quality(M)))
        '{"r_0":   2^35.3,  "r_0/gh": 8.564671,  "rhf": 1.020061,  "/": -0.08317,  "hv/hv": 2.832300}'
        >>> str(pretty_dict(basis_quality(M.r())))
        '{"r_0":   2^35.3,  "r_0/gh": 8.564671,  "rhf": 1.020061,  "/": -0.08317,  "hv/hv": 2.832300}'

    """

    try:
        d = M.d
        r = [M.get_r(i, i) for i in range(d)]
    except AttributeError:
        d = len(M)
        r = M

    ret = OrderedDict()

    log_volume = sum(log(r_)/2 for r_ in r)

    lhs = sum(log(r_)/2 for r_ in r[:d//2])
    rhs = sum(log(r_)/2 for r_ in r[d//2 + (d%2):])

    ret["r_0"] = r[0]
    ret["r_0/gh"] = r[0]/gaussian_heuristic(r)
    ret["rhf"] = exp((log(r[0])/2.0 - log_volume/d)/d)
    try:
        ret['/'] = M.get_current_slope(0, d)
    except AttributeError:
        ret["/"] = get_current_slope(M, 0, d)
    ret["hv/hv"] = exp((lhs - rhs)/d)

    return ret



# -*- coding: utf-8 -*-

from fpylll.fplll.gso import MatGSO
from fpylll.fplll.integer_matrix import IntegerMatrix
from fpylll.fplll.lll import LLLReduction
from fpylll.fplll.enumeration import Enumeration
from fpylll import Pruning
from time import time


def bench_enumeration(n):
    """Return number of nodes visited and wall time for enumeration in dimension `n`.

    :param n: dimension
    :returns: nodes, wall time

    >>> import fpylll.tools.benchmark
    >>> _ = fpylll.tools.benchmark.bench_enumeration(30)

    """

    A = IntegerMatrix.random(n, "qary", bits=30, k=n//2)
    M = MatGSO(A)
    L = LLLReduction(M)
    L(0, 0, n)

    radius = M.get_r(0, 0) * .999
    pruning = Pruning.run(radius, 2.0**50, M.r(), 0.2)

    enum = Enumeration(M)
    t = time()
    enum.enumerate(0, n, radius, 0, pruning=pruning.coefficients)
    t = time() - t
    cost = enum.get_nodes()

    return cost, t



# -*- coding: utf-8 -*-
"""
Plot `\\log_2` of square Gram-Schmidt norms during a BKZ run.

EXAMPLE::

    >>> from fpylll import IntegerMatrix, BKZ, FPLLL
    >>> from fpylll.algorithms.bkz2 import BKZReduction as BKZ2
    >>> from fpylll.tools.bkz_plot import KeepGSOBKZFactory
    >>> FPLLL.set_random_seed(1337)
    >>> _ = FPLLL.set_threads(1)  # to make it deterministic
    >>> A = IntegerMatrix.random(80, "qary", k=40, bits=20)
    >>> bkz = KeepGSOBKZFactory(BKZ2)(A)
    >>> bkz(BKZ.EasyParam(20))
    >>> bkz._KeepGSOBKZ__gso_norms[0][0]
    23104295.0
    >>> bkz._KeepGSOBKZ__gso_norms[-1][0]
    6591824.0

.. modulauthor: Martin Albrecht <martin.albrecht@royalholloway.ac.uk>

"""


def KeepGSOBKZFactory(cls):
    """
    Return a wrapper class around ``cls`` which collects Gram-Schmidt norms in the attribute
    ``__KeepGSOBKZ_gso_norms``.

    In particular, the list will be constructed as follows:

    - index 0: input GSO norms
    - index i,j: kappa and GSO norms in tour i-1 for after j-th SVP call
    - index -1: output GSO norms

    :param cls: A BKZ-like algorithm with methods ``__call__``, ``svp_reduction`` and ``tour``.

    .. warning:: This will slow down the algorithm especially for small block sizes.

    """
    class KeepGSOBKZ(cls):
        def __call__(self, *args, **kwds):
            self.M.update_gso()
            self.__gso_norms = [self.M.r()]
            self.__at_toplevel = True
            cls.__call__(self, *args, **kwds)
            self.M.update_gso()
            self.__gso_norms.append(self.M.r())

        def svp_reduction(self, kappa, *args, **kwds):
            at_toplevel = self.__at_toplevel
            self.__at_toplevel = False
            r = cls.svp_reduction(self, kappa, *args, **kwds)
            self.__at_toplevel = at_toplevel
            if at_toplevel:
                self.M.update_gso()
                self.__gso_norms[-1].append((kappa, self.M.r()))
            return r

        def tour(self, *args, **kwds):
            if self.__at_toplevel:
                self.__gso_norms.append([])
            return cls.tour(self, *args, **kwds)

    return KeepGSOBKZ


def plot_gso_norms(gso_norms, block_size, basename="bkz-gso-norms",
                   extension="png", dpi=300):
    """Plot ``gso_norms``.

    :param gso_norms: list of GSO norms. It is assumed these follow the form output by ``KeepGSOBKZ``.
    :param block_size: BKZ block size
    :param basename: graphics filename basenname (may contain full path)
    :param extension: graphics filename extension/type
    :param dpi: resolution

    :returns: Tuple of filenames written.

    .. note:: To convert to movie, call e.g. ``ffmpeg -framerate 8 -pattern_type glob -i "*.png" bkz.mkv``

    .. warning:: This function is quite slow.
    """
    from math import log, pi, e
    import matplotlib.pyplot as plt
    import matplotlib.patches as patches

    filenames = []

    def maplog2(l):
        return [log(l[i], 2) for i in range(len(l))]

    def plot_finalize(ax, name):
        ax.set_ylabel("$2\\,\\log_2(\\cdot)$")
        ax.set_xlabel("$i$")
        ax.legend(loc="upper right")
        ax.set_ylim(*ylim)

        fullname = "%s.%s"%(name, extension)
        fig.savefig(fullname, dpi=dpi)
        filenames.append(fullname)
        plt.close()

    d = len(gso_norms[0])
    x = range(d)

    beta = float(block_size)
    delta_0 = (beta/(2.*pi*e) * (pi*beta)**(1./beta))**(1./(2.*(beta-1)))
    alpha = delta_0**(-2.*d/(d-1.))
    logvol = sum(maplog2(gso_norms[0]))  # already squared
    gsa = [log(alpha, 2)*(2*i) + log(delta_0, 2)*(2*d) + logvol*(1./d) for i in range(d)]

    fig, ax = plt.subplots()
    ax.plot(x, maplog2(gso_norms[0]), label="$\\|\\mathbf{b}_i^*\\|$")
    ylim = ax.get_ylim()
    ax.set_title("Input")
    plot_finalize(ax, "%s-aaaa-input"%basename)

    for i, tour in enumerate(gso_norms[1:-1]):
        for j, (kappa, norms) in enumerate(tour):
            fig, ax = plt.subplots()

            rect = patches.Rectangle((kappa, ylim[0]), min(block_size, d-kappa-1), ylim[1]-ylim[0],
                                     fill=True, color="lightgray")
            ax.add_patch(rect)
            ax.plot(x, maplog2(norms), label="$\\|\\mathbf{b}_i^*\\|$")
            ax.plot(x, gsa, color="black", label="GSA")
            ax.set_title("BKZ-%d tour: %2d, $\\kappa$: %3d"%(block_size, i, kappa))
            plot_finalize(ax, "%s-t%03d-%04d"%(basename, i, j))

    fig, ax = plt.subplots()
    ax.plot(x, maplog2(gso_norms[-1]))
    ax.set_title("Output")
    plot_finalize(ax, "%s-zzzz-output"%basename)

    return tuple(filenames)



# -*- coding: utf-8 -*-
"""
BKZ 2.0 algorithm in Python.

..  moduleauthor:: Martin R.  Albrecht <martinralbrecht+fpylll@googlemail.com>

"""

from fpylll import BKZ, Enumeration, EnumerationError
from fpylll.algorithms.bkz import BKZReduction as BKZBase
from fpylll.tools.bkz_stats import dummy_tracer
from fpylll.util import gaussian_heuristic, randint


class BKZReduction(BKZBase):

    def __init__(self, A):
        """Create new BKZ object.

        :param A: an integer matrix, a GSO object or an LLL object

        """
        BKZBase.__init__(self, A)

    def get_pruning(self, kappa, block_size, params, tracer=dummy_tracer):
        strategy = params.strategies[block_size]
        radius, re = self.M.get_r_exp(kappa, kappa)
        radius *= self.lll_obj.delta
        r = [self.M.get_r_exp(i, i) for i in range(kappa, kappa+block_size)]
        gh_radius = gaussian_heuristic([x for x, _ in r])
        ge = float(sum([y for _, y in r])) / len(r)

        if (params.flags & BKZ.GH_BND and block_size > 30):
            radius = min(radius, gh_radius * 2**(ge-re) * params.gh_factor)

        return radius, re, strategy.get_pruning(radius, gh_radius * 2**(ge-re))

    def randomize_block(self, min_row, max_row, tracer=dummy_tracer, density=0):
        """Randomize basis between from ``min_row`` and ``max_row`` (exclusive)

            1. permute rows

            2. apply lower triangular matrix with coefficients in -1,0,1

        :param min_row: start in this row
        :param max_row: stop at this row (exclusive)
        :param tracer: object for maintaining statistics
        :param density: number of non-zero coefficients in lower triangular transformation matrix
        """
        if max_row - min_row < 2:
            return  # there is nothing to do

        # 1. permute rows
        niter = 4 * (max_row-min_row)  # some guestimate
        with self.M.row_ops(min_row, max_row):
            for i in range(niter):
                b = a = randint(min_row, max_row-1)
                while b == a:
                    b = randint(min_row, max_row-1)
                self.M.move_row(b, a)

        # 2. triangular transformation matrix with coefficients in -1,0,1
        with self.M.row_ops(min_row, max_row):
            for a in range(min_row, max_row-2):
                for i in range(density):
                    b = randint(a+1, max_row-1)
                    s = randint(0, 1)
                    self.M.row_addmul(a, b, 2*s-1)

        return

    def svp_preprocessing(self, kappa, block_size, params, tracer=dummy_tracer):
        clean = True

        clean &= BKZBase.svp_preprocessing(self, kappa, block_size, params, tracer)

        for preproc in params.strategies[block_size].preprocessing_block_sizes:
            prepar = params.__class__(block_size=preproc, strategies=params.strategies, flags=BKZ.GH_BND)
            clean &= self.tour(prepar, kappa, kappa + block_size, tracer=tracer)

        return clean

    def svp_reduction(self, kappa, block_size, params, tracer=dummy_tracer):
        """

        :param kappa:
        :param block_size:
        :param params:
        :param tracer:

        """

        self.lll_obj.size_reduction(0, kappa+1)
        old_first, old_first_expo = self.M.get_r_exp(kappa, kappa)

        remaining_probability, rerandomize = 1.0, False

        while remaining_probability > 1. - params.min_success_probability:
            with tracer.context("preprocessing"):
                if rerandomize:
                    with tracer.context("randomization"):
                        self.randomize_block(kappa+1, kappa+block_size,
                                             density=params.rerandomization_density, tracer=tracer)
                with tracer.context("reduction"):
                    self.svp_preprocessing(kappa, block_size, params, tracer=tracer)

            with tracer.context("pruner"):
                radius, re, pruning = self.get_pruning(kappa, block_size, params, tracer)

            try:
                enum_obj = Enumeration(self.M)
                with tracer.context("enumeration",
                                    enum_obj=enum_obj,
                                    probability=pruning.expectation,
                                    full=block_size==params.block_size):
                    max_dist, solution = enum_obj.enumerate(kappa, kappa + block_size, radius, re,
                                                            pruning=pruning.coefficients)[0]
                with tracer.context("postprocessing"):
                    self.svp_postprocessing(kappa, block_size, solution, tracer=tracer)
                rerandomize = False

            except EnumerationError:
                rerandomize = True

            remaining_probability *= (1 - pruning.expectation)

        self.lll_obj.size_reduction(0, kappa+1)
        new_first, new_first_expo = self.M.get_r_exp(kappa, kappa)

        clean = old_first <= new_first * 2**(new_first_expo - old_first_expo)
        return clean



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from fpylll.algorithms.simple_bkz import BKZReduction
from fpylll import Enumeration
import math
from functools import reduce


class DBKZReduction(BKZReduction):
    def bkz_loop(self, block_size, min_row, max_row):
        """FIXME! briefly describe function

        :param block_size:
        :param min_row:
        :param max_row:
        :returns:
        :rtype:

        """
        self.m.update_gso()
        clean = True
        for kappa in range(max_row - block_size, min_row - 1, -1):
            clean &= self.dsvp_reduction(kappa, block_size)
        clean &= BKZReduction.bkz_loop(self, block_size, min_row, max_row)
        return clean

    def euclid(self, pair1, pair2):
        """FIXME! briefly describe function

        :param pair1:
        :param pair2:
        :returns:
        :rtype:

        """
        row1, x1 = pair1
        row2, x2 = pair2
        if not x1:
            return pair2
        c = math.floor(x2/x1)
        self.m.row_addmul(row2, row1, -c)
        return self.euclid((row2, x2 - c*x1), pair1)

    def dsvp_reduction(self, kappa, block_size):
        """FIXME! briefly describe function

        :param kappa:
        :param block_size:
        :returns:
        :rtype:

        """
        clean = True

        self.lll_obj(0, kappa, kappa + block_size)
        if self.lll_obj.nswaps > 0:
            clean = False

        max_dist, expo = self.m.get_r_exp(kappa + block_size - 1, kappa + block_size - 1)
        max_dist = 1.0/max_dist
        expo *= -1.0
        delta_max_dist = self.lll_obj.delta * max_dist

        max_dist, solution = Enumeration(self.m).enumerate(kappa, kappa + block_size, max_dist, expo,
                                                           pruning=None, dual=True)[0]
        if max_dist >= delta_max_dist:
            return clean

        with self.m.row_ops(kappa, kappa+block_size):
            pairs = list(enumerate(solution, start=kappa))
            [self.m.negate_row(pair[0]) for pair in pairs if pair[1] < 0]
            pairs = map(lambda x: (x[0], abs(x[1])), pairs)
            # GCD should be tree based but for proof of concept implementation, this will do
            row, x = reduce(self.euclid, pairs)
            if x != 1:
                raise RuntimeError("Euclid failed!")
            self.m.move_row(row, kappa + block_size - 1)
        self.lll_obj(kappa, kappa, kappa + block_size)

        return False



# -*- coding: utf-8 -*-
"""
Compare the performance of BKZ variants.

..  moduleauthor:: Martin R.  Albrecht <martinralbrecht+fpylll@googlemail.com>

..  note :: This module does not work with standard BKZ classes.  Instead, it expects that the BKZ
classes it consumes accept a block size as ``params`` in the ``tour`` member function.  This way,
the construction of param objects can be rolled into the class description. See example classes below.
"""


# Imports

from __future__ import absolute_import
from collections import OrderedDict
from fpylll import IntegerMatrix, BKZ
from fpylll import FPLLL
from fpylll.tools.bkz_stats import BKZTreeTracer, dummy_tracer, pretty_dict
from fpylll.tools.quality import basis_quality
from fpylll.util import ReductionError

from multiprocessing import Pool
import logging
import copy
import time
import pickle

import fpylll.algorithms.bkz
import fpylll.algorithms.bkz2


# Utility Functions

def play(BKZ, A, block_size, tours, progressive_step_size=None):
    """Call ``BKZ`` on ``A`` with ``block_size`` for the given number of ``tours``.

    The given number of tours is used for all block sizes from 2 up to ``block_size`` in increments of
    ``progressive_step_size``. Providing ``None`` for this parameter disables the progressive strategy.

    :param BKZ: a BKZ class whose ``__call__`` accepts a single block size as parameter
    :param A: an integer matrix
    :param block_size: a block size >= 2
    :param tours: number of tours >= 1
    :param progressive_step_size: step size for progressive strategy
    :returns: a trace of the execution using ``BKZTreeTracer``

    ..  note :: This function essentially reimplements ``BKZ.__call__`` but supports the
        progressive strategy.
    """
    bkz = BKZ(copy.copy(A))
    tracer = BKZTreeTracer(bkz, start_clocks=True)

    # this essentially initialises the GSO object, LLL was already run by the constructor, so this
    # is quick.
    with tracer.context("lll"):
        bkz.lll_obj()

    if progressive_step_size is None:
        block_sizes = (block_size,)
    elif int(progressive_step_size) > 0:
        block_sizes = range(2, block_size+1, progressive_step_size)
        if block_sizes[-1] != block_size:
            block_sizes.append(block_size)
    else:
        raise ValueError("Progressive step size of %s not understood."%progressive_step_size)

    for block_size in block_sizes:
        for i in range(tours):
            with tracer.context("tour", (block_size, i)):
                bkz.tour(block_size, tracer=tracer)
    tracer.exit()
    trace = tracer.trace

    quality = basis_quality(bkz.M)
    for k, v in quality.items():
        trace.data[k] = v

    return trace


class Conductor(object):
    """
    A conductor is our main class for launching block-wise lattice reductions and collecting the outputs.
    """
    def __init__(self, threads=1, pickle_jar=None, logger="."):
        """Create a new conductor object.

        :param threads: number of threads
        :param pickle_jar: dump traces to this file continuously

        """
        self.pool = Pool(processes=threads)
        self.threads = threads
        self.pickle_jar = pickle_jar
        self.logger = logging.getLogger(logger)
        self.outputs = OrderedDict()
        self._major_strlen = 0
        self._minor_strlen = 0

    def _majorminor_format_str(self):
        "Used to align log files"
        return "%%%ds(%%%ds) :: %%s"%(self._major_strlen, self._minor_strlen)

    def _update_strlens(self, major, minor):
        "Update string lengths of major/minor tags"
        self._major_strlen = max(len(str(major)), self._major_strlen)
        self._minor_strlen = max(len(str(minor)), self._minor_strlen)

    @staticmethod
    def dump(data, filename):
        "Pickle ``data`` to ``filename``"
        pickle.dump(data, open(filename, "wb"))

    def wait_on(self, outputs, todo, sleep=1):
        """Wait for jobs in ``todo`` to return and store results in ``outputs``.

        :param outputs: store results here
        :param todo: these are running jobs
        :param sleep: seconds to sleep before checking if new results are availabl.

        """

        fmtstr = self._majorminor_format_str()

        while todo:
            collect = [(tag, res) for (tag, res) in todo if res.ready()]

            for tag, res in collect:
                major, minor = tag
                try:
                    res = res.get()
                    if major not in outputs:
                        outputs[major] = []
                    outputs[major].append((minor, res))
                    self.logger.debug(fmtstr%(major, minor, pretty_dict(res.data)))

                    if self.pickle_jar is not None:
                        Conductor.dump(self.outputs, self.pickle_jar)

                except ReductionError:
                    self.logger.debug("ReductionError for %s(%s)."%(major, minor))

            todo = todo.difference(collect)
            time.sleep(sleep)

        return outputs

    def log_averages(self, tags, outputs):
        """
        Log average values for all entries tagged as ``tags`` in ``outputs``.
        """
        fmtstr = self._majorminor_format_str()
        avg = OrderedDict()

        for major, minor in tags:
            if major in avg:
                continue
            avg[major] = OrderedDict()
            n = len(outputs[major])
            for minor, output in outputs[major]:
                for k, v in output.data.items():
                    avg[major][k] =  avg[major].get(k, 0.0) + float(v)/n

            self.logger.info(fmtstr%(major, "avg", pretty_dict(avg[major])))

    def __call__(self, jobs, current=None):
        """
        Call ``jobs`` in parallel.

        The parameter jobs is a list with the following format.  Each entry is one of the following:

            - a tuple ``((major, minor), (BKZ, A, block_size, tours, progressive_step_size))``,
              where ``major`` and ``minor`` are arbitrary hashable tags and the rest are valid
              inputs to ``play``.

            - A list with elements of the same format as above.

        Entries at the same level are considered to be a group.  All jobs in the same group go into
        the same execution pool.  At the end of the execution of a group the average across all
        ``minor`` tags of a ``major`` tag are shown.

        ..  note :: Recursive jobs, i.e. those in a sub-list are run first, this is an
            implementation artefact.  Typically, we don't expect jobs and lists of jobs to be mixed at
            the same level, though, this is supported.

        """
        inputs = OrderedDict()
        if current is None:
            current = self.outputs

        # filter out sub-jobs that should be grouped and call recursively
        for tag, job in jobs:
            if isinstance(job[0], (list, tuple)):
                self.logger.info("")
                self.logger.info("# %s (size: %d) #"%(tag, len(job)))
                current[tag] = OrderedDict()
                self(job, current=current[tag])
            else:
                major, minor = tag
                self._update_strlens(major, minor)
                if major not in current:
                    current[major] = list()
                inputs[tag] = job

        self.logger.debug("")

        # base case
        if self.threads > 1:
            todo = set()
            for tag in inputs:
                todo.add((tag, self.pool.apply_async(play, inputs[tag])))

            current = self.wait_on(current, todo)

        else:
            fmtstr = self._majorminor_format_str()
            for tag in inputs:
                major, minor = tag
                try:
                    res = play(*inputs[tag])
                    current[major].append((minor, res))
                    self.logger.debug(fmtstr%(major, minor, pretty_dict(res.data)))

                    if self.pickle_jar is not None:
                        Conductor.dump(self.outputs, self.pickle_jar)

                except ReductionError:
                    self.logger.debug("ReductionError for %s(%s)."%(major, minor))

        self.logger.debug("")

        # print averages per major tag
        self.log_averages(inputs.keys(), current)

        if self.pickle_jar is not None:
            Conductor.dump(self.outputs, self.pickle_jar)

        return self.outputs


def compare_bkz(classes, matrixf, dimensions, block_sizes, progressive_step_size,
                seed, threads=2, samples=2, tours=1,
                pickle_jar=None, logger="compare"):
    """
    Compare BKZ-style lattice reduction.

    :param classes: a list of BKZ classes to test.  See caveat above.
    :param matrixf: A function to create matrices for a given dimension and block size
    :param dimensions: a list of dimensions to test
    :param block_sizes: a list of block sizes to test
    :param progressive_step_size: step size for the progressive strategy; ``None`` to disable it
    :param seed: A random seed, each matrix will be created with seed increased by one
    :param threads: number of threads to use
    :param samples: number of reductions to perform
    :param tours: number of BKZ tours to run
    :param log_filename: log to this file if not ``None``

    """

    jobs = []

    for dimension in dimensions:

        jobs.append((dimension, []))

        for block_size in block_sizes:
            if dimension < block_size:
                continue

            seed_ = seed
            jobs_ = []

            matrixf_ = matrixf(dimension=dimension, block_size=block_size)

            for i in range(samples):
                FPLLL.set_random_seed(seed_)
                A = IntegerMatrix.random(dimension, **matrixf_)

                for BKZ_ in classes:
                    args = (BKZ_, A, block_size, tours, progressive_step_size)
                    jobs_.append(((BKZ_.__name__, seed_), args))
                seed_ += 1

            jobs[-1][1].append((block_size, jobs_))

    conductor = Conductor(threads=threads, pickle_jar=pickle_jar, logger=logger)
    return conductor(jobs)


# Example

class BKZGlue(object):
    "Base class for producing new BKZ classes with some parameters fixed."
    def tour(self, params, min_row=0, max_row=-1, tracer=dummy_tracer):
        if isinstance(params, int):
            params = BKZ.Param(block_size=params, **self.kwds)
        return self.base.tour(self, params, min_row=min_row, max_row=max_row, tracer=tracer)


def BKZFactory(name, BKZBase, **kwds):
    """
    Return a new BKZ class, derived from ``BKZBase`` with given ``name``.  The resulting class
    accepts a single ``block_size`` parameter for ``tour`` and substitutes it with a ``BKZ.Param```
    object where the keyword parameters provided to this function are fixed the values provided to
    this function ::

        >>> from fpylll.algorithms.bkz2 import BKZReduction as BKZ2
        >>> from fpylll import BKZ
        >>> from fpylll.tools.compare import BKZFactory
        >>> BKZ2_LOW = BKZFactory('BKZ2_LOW', BKZ2, strategies=BKZ.DEFAULT_STRATEGY, min_success_probability=0.1)

    :param name: name for output class
    :param BKZBase: base class to base this class on

    """
    NEW_BKZ = type(name, (BKZGlue, BKZBase), {"kwds": kwds, "base": BKZBase})
    globals()[name] = NEW_BKZ  # this is a HACK to enable pickling
    return NEW_BKZ


BKZ1 = BKZFactory("BKZ1", fpylll.algorithms.bkz.BKZReduction)
BKZ2 = BKZFactory("BKZ2", fpylll.algorithms.bkz2.BKZReduction, strategies=BKZ.DEFAULT_STRATEGY)


# Main

def qary30(dimension, block_size):
    return {"algorithm": "qary",
            "k": dimension//2,
            "bits": 30,
            "int_type": "long"}


def setup_logging(name, verbose=False):
    import subprocess
    import datetime

    hostname = str(subprocess.check_output("hostname").rstrip())
    now = datetime.datetime.today().strftime("%Y-%m-%d-%H:%M")
    log_name = "{name}-{hostname}-{now}".format(name=name, hostname=hostname, now=now)

    logging.basicConfig(level=logging.DEBUG,
                        format='%(levelname)5s:%(name)s:%(asctime)s: %(message)s',
                        datefmt='%Y/%m/%d %H:%M:%S %Z',
                        filename=log_name + ".log")

    # define a Handler which writes INFO messages or higher to the sys.stderr
    console = logging.StreamHandler()
    console.setLevel(logging.INFO if not verbose else logging.DEBUG)
    console.setFormatter(logging.Formatter('%(name)s: %(message)s',))
    logging.getLogger(name).addHandler(console)

    return log_name


def names_to_classes(class_names, filenames):
    """
    Try to find a class for each name in ``class_names``.  Classes implemented in one of the
    ``filenames`` are also considered.

    The following mapping logic is used:

        - if a class with ``name`` exists, it is used

        - if ``name`` is ``BKZ_FOO`` and a class called ``BKZReduction`` is implemented in a file
          ``bkz_foo`` it is used.

    :param class_names:
    :param filenames:

    """
    import imp
    import os
    import re

    classes = class_names
    classes = [globals().get(clas, clas) for clas in classes]

    for i, fn in enumerate(filenames):
        tmp = imp.load_source("compare_module%03d"%i, fn)
        # find the class by name in the module
        classes = [tmp.__dict__.get(clas, clas) for clas in classes]

        # check if there's some BKZReduction implemented in bkz_foo.py, we match this with BKZ_FOO
        if fn.startswith(os.path.basename(fn)) and "BKZReduction" in tmp.__dict__:
            candidate = re.sub("bkz(.*)\\.py", "BKZ\\1", os.path.basename(fn)).upper()
            if candidate in classes:
                tmp.BKZReduction.__name__ = candidate
                classes[classes.index(candidate)] = tmp.BKZReduction

    for clas in classes:
        if isinstance(clas, str):
            raise ValueError("Cannot find '%s'"%clas)

    return classes


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(description='Measure Pressure')
    parser.add_argument('-c', '--classes', help='BKZ classes',
                        type=str, nargs='+', default=["BKZ2"])
    parser.add_argument('-f', '--files', help='additional files to load for BKZ classes',
                        type=str, nargs='+', default=list())
    parser.add_argument('-t', '--threads', help='number of threads to use', type=int, default=1)
    parser.add_argument('-r', '--tours',   help='number of BKZ tours', type=int, default=1)
    parser.add_argument('-s', '--samples', help='number of samples to try', type=int, default=4)
    parser.add_argument('-z', '--seed', help="random seed", type=int, default=0x1337)
    parser.add_argument('-b', '--block-sizes', help='block sizes',
                        type=int,  nargs='+', default=(10, 20, 30, 40))
    parser.add_argument('-p', '--progressive-step-size', help='step size for progressive strategy, None for disabled',
                        default=None, type=int)
    parser.add_argument('-d', '--dimensions', help='lattice dimensions',
                        type=int, nargs='+', default=(60, 80, 100, 120))
    parser.add_argument('-v', '--verbose', help='print more details by default', action='store_true')

    args =  parser.parse_args()

    name = "compare"

    classes = names_to_classes(args.classes, args.files)
    log_filename = setup_logging(name, args.verbose)

    for k, v in sorted(vars(args).items()):
        logging.getLogger(name).debug("%s: %s"%(k, v))

    results = compare_bkz(classes=classes,
                          matrixf=qary30,
                          block_sizes=args.block_sizes,
                          progressive_step_size=args.progressive_step_size,
                          dimensions=args.dimensions,
                          logger=name,
                          pickle_jar=log_filename + ".sobj",
                          seed=args.seed,
                          threads=args.threads,
                          samples=args.samples,
                          tours=args.tours)



# flake8: noqa
from __future__ import absolute_import
from .fplll.integer_matrix import IntegerMatrix
from .fplll.gso import GSO
from .fplll.lll import LLL
from .fplll.enumeration import Enumeration, EnumerationError, EvaluatorStrategy
from .fplll.bkz import BKZ
from .fplll.bkz_param import load_strategies_json
from .fplll.svpcvp import SVP
from .fplll.svpcvp import CVP
from .fplll.pruner import Pruning
from .util import ReductionError
from .util import FPLLL

__version__ = "0.6.4"



# -*- coding: utf-8 -*-
"""
A minimal implementation of the Block Korkine Zolotarev algorithm in Python.

..  moduleauthor:: Martin R.  Albrecht <martinralbrecht+fpylll@googlemail.com>
"""

from __future__ import absolute_import
from fpylll import IntegerMatrix, GSO, LLL, BKZ
from fpylll import Enumeration


class BKZReduction:
    def __init__(self, A):
        """Construct a new BKZ reduction instance.

        :param A: Integer matrix to reduce.

        """
        if not isinstance(A, IntegerMatrix):
            raise TypeError("Matrix must be IntegerMatrix but got type '%s'"%type(A))

        # run LLL first
        wrapper = LLL.Wrapper(A)
        wrapper()

        self.A = A
        self.m = GSO.Mat(A, flags=GSO.ROW_EXPO)
        self.lll_obj = LLL.Reduction(self.m)

    def __call__(self, block_size):
        """Perform BKZ reduction with given``block_size``.

        Nothing is returned, the matrix ``A`` given during construction is modified in-place.

        :param block_size: an integer > 2

        """
        self.m.discover_all_rows()

        auto_abort = BKZ.AutoAbort(self.m, self.A.nrows)

        while True:
            clean = self.bkz_loop(block_size, 0, self.A.nrows)
            if clean:
                break
            if auto_abort.test_abort():
                break

    def bkz_loop(self, block_size, min_row, max_row):
        """Perform one BKZ loop, often also called a "BKZ tour".

        :param block_size: an integer > 2
        :param min_row: algorithm starts in this row (inclusive)
        :param max_row: algorithm stops at this row (exclusive)

        """
        clean = True
        for kappa in range(min_row, max_row-1):
            bs = min(block_size, max_row - kappa)
            clean &= self.svp_reduction(kappa, bs)
        return clean

    def svp_reduction(self, kappa, block_size):
        """Call the SVP oracle and insert found vector into basis.

        :param kappa: row index
        :param block_size: an integer > 2

        """
        clean = True

        self.lll_obj(0, kappa, kappa + block_size)
        if self.lll_obj.nswaps > 0:
            clean = False

        max_dist, expo = self.m.get_r_exp(kappa, kappa)
        delta_max_dist = self.lll_obj.delta * max_dist

        max_dist, solution = Enumeration(self.m).enumerate(kappa, kappa + block_size, max_dist, expo, pruning=None)[0]

        if max_dist >= delta_max_dist * (1<<expo):
            return clean

        nonzero_vectors = len([x for x in solution if x])

        if nonzero_vectors == 1:
            first_nonzero_vector = None
            for i in range(block_size):
                if abs(solution[i]) == 1:
                    first_nonzero_vector = i
                    break

            self.m.move_row(kappa + first_nonzero_vector, kappa)
            self.lll_obj.size_reduction(kappa, kappa + first_nonzero_vector + 1)

        else:
            d = self.m.d
            self.m.create_row()

            with self.m.row_ops(d, d+1):
                for i in range(block_size):
                    self.m.row_addmul(d, kappa + i, solution[i])

            self.m.move_row(d, kappa)
            self.lll_obj(kappa, kappa, kappa + block_size + 1)
            self.m.move_row(kappa + block_size, d)

            self.m.remove_last_row()

        return False
